"""Model for `/api/users/:id/unread_messages`."""
from pydantic import BaseModel


class UnreadMessages(BaseModel):
    """Represents counter entity for unread messages/news/notifications."""
    messages: int
    news: int
    notifications: int



"""Model for `/api/dialogs`."""
from pydantic import BaseModel

from .message import MessageInfo
from .user import UserInfo


class Dialog(BaseModel):
    """Represents dialog entity."""
    target_user: UserInfo
    message: MessageInfo



"""Enums for `/api/animes/:anime_id/videos`."""
from .enhanced_enum import EnhancedEnum


class VideoKind(EnhancedEnum):
    """Contains constants related for video kind."""
    PV = 'pv'
    CHARACTER_TRAILER = 'character_trailer'
    CM = 'cm'
    OP = 'op'
    ED = 'ed'
    OP_ED_CLIP = 'op_ed_clip'
    CLIP = 'clip'
    OTHER = 'other'
    EPISODE_PREVIEW = 'episode_preview'



"""Represents `/api/constants` resource."""
from typing import Any, Dict, List, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import (AnimeConstants, ClubConstants, MangaConstants,
                      SmileyConstant, UserRateConstants)
from ..utils import Utils
from .base_resource import BaseResource


class Constants(BaseResource):
    """Constants resource class.

    Used to represent `/api/constants` resource
    """

    @method_endpoint('/api/constants/anime')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def anime(self):
        """Returns anime constants values.

        :return: Anime constants values
        :rtype: Optional[AnimeConstants]
        """
        response = await self._client.request(
            self._client.endpoints.anime_constants)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=AnimeConstants)

    @method_endpoint('/api/constants/manga')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def manga(self):
        """Returns manga constants values.

        :return: Manga constants values
        :rtype: Optional[MangaConstants]
        """
        response = await self._client.request(
            self._client.endpoints.manga_constants)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=MangaConstants)

    @method_endpoint('/api/constants/user_rate')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def user_rate(self):
        """Returns user rate constants values.

        :return: User rate constants values
        :rtype: Optional[UserRateConstants]
        """
        response = await self._client.request(
            self._client.endpoints.user_rate_constants)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=UserRateConstants)

    @method_endpoint('/api/constants/club')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def club(self):
        """Returns club constants values.

        :return: Club constants values
        :rtype: Optional[ClubConstants]
        """
        response = await self._client.request(
            self._client.endpoints.club_constants)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=ClubConstants)

    @method_endpoint('/api/constants/smileys')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def smileys(self):
        """Returns list of smiley constant values.

        :return: List of smiley constant values
        :rtype: List[SmileyConstant]
        """
        response = await self._client.request(
            self._client.endpoints.smileys_constants)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=SmileyConstant)



"""General exception for raising on other errors."""

from loguru import logger


class ShikithonException(Exception):

    def __init__(self, error_message: str):
        logger.error(error_message)
        super().__init__(error_message)



"""Model for `/api/user_rates` and `/api/v2/user_rates`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class UserRate(BaseModel):
    """Represents user rate entity."""
    id: int
    user_id: Optional[int]
    target_id: Optional[int]
    target_type: Optional[str]
    score: int
    status: str
    text: Optional[str]
    episodes: Optional[int]
    chapters: Optional[int]
    volumes: Optional[int]
    text_html: str
    rewatches: int
    created_at: datetime
    updated_at: datetime



"""Enums for `/api/users/:id/history`."""
from .enhanced_enum import EnhancedEnum


class HistoryTargetType(EnhancedEnum):
    """Contains constants related for history target type."""
    ANIME = 'Anime'
    MANGA = 'Manga'



"""Enums for `/api/styles`."""
from .enhanced_enum import EnhancedEnum


class StyleOwner(EnhancedEnum):
    """Contains constants related for style owner type."""
    USER = 'User'
    CLUB = 'Club'



"""Exception for raising on missing app variable."""
from typing import List, Union

from loguru import logger


class MissingAppVariable(Exception):
    """Exception for raising on missing app variable."""

    def __init__(self, variable_name: Union[str, List[str]]):
        missing_variables = 'Missing app variable(s): '

        if isinstance(variable_name, list):
            missing_variables += ', '.join(variable_name)
        else:
            missing_variables += variable_name

        error_message = 'There is a problem with some app variables. ' \
              'Recheck your config and try again.\n' \
              f'{missing_variables}'
        logger.error(error_message)
        super().__init__(error_message)



"""Model for `/api/styles`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class Style(BaseModel):
    """Represents style entity.

    Many fields are optional due to
    `/api/styles/preview` returning non-null
    for 3 fields only
    """
    id: Optional[int]
    owner_id: Optional[int]
    owner_type: Optional[str]
    name: str
    css: str
    compiled_css: Optional[str]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]



"""Submodel for `people.py`."""
from typing import List

from pydantic import BaseModel

from .anime import AnimeInfo
from .character import CharacterInfo


class Roles(BaseModel):
    """Represents roles entity of person."""
    characters: List[CharacterInfo]
    animes: List[AnimeInfo]



"""Represents `/api/styles` resource."""
from typing import Any, Dict, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType, StyleOwner
from ..exceptions import ShikimoriAPIResponseError
from ..models import Style
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'style'


class Styles(BaseResource):
    """Styles resource class.

    Used to represent `/api/styles` resource
    """

    @method_endpoint('/api/styles/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, style_id: int):
        """Returns info about style.

        :param style_id: Style ID to get info
        :type style_id: int

        :return: Info about style
        :rtype: Optional[Style]
        """
        response = await self._client.request(
            self._client.endpoints.style(style_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Style)

    @method_endpoint('/api/styles/preview')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def preview(self, css: str):
        """Previews style with passed CSS code.

        :param css: CSS code to preview
        :type css: str

        :return: Info about previewed style
        :rtype: Optional[Style]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME, css=css)

        response = await self._client.request(
            self._client.endpoints.style_preview,
            data=data_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Style)

    @method_endpoint('/api/styles')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self, css: str, name: str, owner_id: int,
                     owner_type: StyleOwner):
        """Creates new style.

        :param css: CSS code for style
        :type css: str

        :param name: Style name
        :type name: str

        :param owner_id: User/Club ID for style ownership
        :type owner_id: int

        :param owner_type: Type of owner
        :type owner_type: StyleOwner

        :return: Info about previewed style
        :rtype: Optional[Style]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           css=css,
                                           name=name,
                                           owner_id=owner_id,
                                           owner_type=owner_type)

        response = await self._client.request(self._client.endpoints.styles,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Style)

    @method_endpoint('/api/styles/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self,
                     style_id: int,
                     css: Optional[str] = None,
                     name: Optional[str] = None):
        """Updates existing style.

        :param style_id: ID of existing style for edit
        :type style_id: int

        :param css: New CSS code for style
        :type css: Optional[str]

        :param name: New style name
        :type name: Optional[str]

        :return: Info about updated style
        :rtype: Optional[Style]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           css=css,
                                           name=name)

        response = await self._client.request(
            self._client.endpoints.style(style_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Style)



"""Represents `/api/favorites` resource."""
from typing import Any, Dict, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import FavoriteLinkedType, PersonKind, RequestType, ResponseCode
from ..exceptions import ShikimoriAPIResponseError
from ..utils import Utils
from .base_resource import BaseResource


class Favorites(BaseResource):
    """Favorites resource class.

    Used to represent `/api/favorites` resource
    """

    @method_endpoint('/api/favorites/:linked_type/:linked_id(/:kind)')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def create(self,
                     linked_type: FavoriteLinkedType,
                     linked_id: int,
                     kind: Optional[PersonKind] = None):
        """Creates a favorite.

        :param linked_type: Type of object for making favorite
        :type linked_type: FavoriteLinkedType

        :param linked_id: ID of linked type
        :type linked_id: int

        :param kind: Kind of linked type
            (Required when linked_type is 'Person')
        :type kind: Optional[PersonKind]

        :return: Status of favorite create
        :rtype: bool
        """
        linked_type_value = str(linked_type)
        kind_value = None if kind is None else str(kind)

        response = await self._client.request(
            self._client.endpoints.favorites_create(linked_type_value,
                                                    linked_id, kind_value),
            request_type=RequestType.POST)

        return cast(Dict[str, Any], response).get('success') is True

    @method_endpoint('/api/favorites/:linked_type/:linked_id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def destroy(self, linked_type: FavoriteLinkedType, linked_id: int):
        """Destroys a favorite.

        :param linked_type: Type of object for destroying from favorite
        :type linked_type: FavoriteLinkedType

        :param linked_id: ID of linked type
        :type linked_id: int

        :return: Status of favorite destroy
        :rtype: bool
        """
        linked_type_value = str(linked_type)

        response = await self._client.request(
            self._client.endpoints.favorites_destroy(linked_type_value,
                                                     linked_id),
            request_type=RequestType.DELETE)

        return cast(Dict[str, Any], response).get('success') is True

    @method_endpoint('/api/favorites/:id/reorder')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def reorder(self, favorite_id: int, new_index: Optional[int] = None):
        """Reorders a favorite to the new index.

        This method requires a favorite ID,
        which cannot be retrieved through the current API methods

        To get the favorite ID, use DevTools on the favorites page
        See https://github.com/shikimori/shikimori/issues/2655

        :param favorite_id: ID of a favorite to reorder
        :type favorite_id: int

        :param new_index: Index of a new position of favorite.
            If skipped, sets favorite to the first position
        :type new_index: Optional[int]

        :return: Status of reorder
        :rtype: bool
        """
        query_dict = Utils.create_query_dict(new_index=new_index)

        response = await self._client.request(
            self._client.endpoints.favorites_reorder(favorite_id),
            query=query_dict,
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)



"""Model for `/api/constants`."""
from typing import Tuple, Literal

from pydantic import BaseModel


class AnimeConstants(BaseModel):
    """Represents anime constants."""
    kind: Tuple[Literal['tv'], Literal['movie'], Literal['ova'], Literal['ona'],
                Literal['special'], Literal['music']]
    status: Tuple[Literal['anons'], Literal['ongoing'], Literal['released']]


class MangaConstants(BaseModel):
    """Represents manga constants."""
    kind: Tuple[Literal['manga'], Literal['manhwa'], Literal['manhua'],
                Literal['light_novel'], Literal['novel'], Literal['one_shot'],
                Literal['doujin']]
    status: Tuple[Literal['anons'], Literal['ongoing'], Literal['released'],
                  Literal['paused'], Literal['discontinued']]


class UserRateConstants(BaseModel):
    """Represents user rate constants."""
    status: Tuple[Literal['planned'], Literal['watching'],
                  Literal['rewatching'], Literal['completed'],
                  Literal['on_hold'], Literal['dropped']]


class ClubConstants(BaseModel):
    """Represents clubs constants."""
    join_policy: Tuple[
        Literal['free'],
        Literal['member_invite'],
        Literal['admin_invite'],
        Literal['owner_invite'],
    ]
    comment_policy: Tuple[
        Literal['free'],
        Literal['members'],
        Literal['admins'],
    ]
    image_upload_policy: Tuple[
        Literal['members'],
        Literal['admins'],
    ]


class SmileyConstant(BaseModel):
    """Represents smiley constant."""
    bbcode: str
    path: str



"""Submodel for `message.py`."""
from datetime import date
from typing import Optional

from pydantic import BaseModel

from .image import Image


class LinkedTopic(BaseModel):
    """Represents linked topic of message entity."""
    id: int
    topic_url: str
    thread_id: int
    topic_id: int
    type: str
    name: Optional[str]
    russian: Optional[str]
    image: Optional[Image]
    url: Optional[str]
    kind: Optional[str]
    score: Optional[float]
    status: Optional[str]
    episodes: Optional[int]
    episodes_aired: Optional[int]
    aired_on: Optional[date]
    released_on: Optional[date]



"""Decorator for logging method endpoint."""
from functools import wraps
from typing import Awaitable, Callable, TypeVar

from loguru import logger
from typing_extensions import ParamSpec

P = ParamSpec('P')
R = TypeVar('R')


def method_endpoint(
    method_endpoint_name: str
) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
    """Decorator for logging method endpoint.

    :param method_endpoint_name: Name of method endpoint
    :type method_endpoint_name: str

    :return: Decorator function
    :rtype: Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]
    """

    def endpoint_logger_wrapper(
            function: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        """Endpoint logger wrapper.

        :param function: Function to decorate
        :type function: Callable[P, Awaitable[R]]

        :return: Decorated function
        :rtype: Callable[P, Awaitable[R]]
        """

        @wraps(function)
        async def endpoint_logger_wrapped(*args: P.args,
                                          **kwargs: P.kwargs) -> R:
            """Decorator's wrapped function for logging endpoint of method.

            :param args: Positional arguments
            :type args: P.args

            :param kwargs: Keyword arguments
            :type kwargs: P.kwargs

            :return: Result of decorated function
            :rtype: R
            """
            logger.debug(f'Executing "{method_endpoint_name}" method')
            return await function(*args, **kwargs)

        return endpoint_logger_wrapped

    return endpoint_logger_wrapper



"""Represents `/api/achievements` resource."""

from typing import Any, Dict, List, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import Achievement
from ..utils import Utils
from .base_resource import BaseResource


class Achievements(BaseResource):
    """Achievements resource class.

    Used to represent `/api/achievements` resource
    """

    @method_endpoint('/api/achievements')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self, user_id: int):
        """Returns list of user achievements.

        :param user_id: User ID for getting achievements
        :type user_id: int

        :return: List of achievements
        :rtype: List[Achievement]
        """
        query_dict = Utils.create_query_dict(user_id=user_id)

        response = await self._client.request(
            self._client.endpoints.achievements, query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Achievement)



"""Enums for `/api/person` and `/api/favorites`."""
from .enhanced_enum import EnhancedEnum


class PersonKind(EnhancedEnum):
    """Contains constants related for favorite person kind."""
    COMMON = 'common'
    SEYU = 'seyu'
    MANGAKA = 'mangaka'
    PRODUCER = 'producer'
    PERSON = 'person'


class PersonSearchKind(EnhancedEnum):
    """Contains constants related for search person kind."""
    SEYU = 'seyu'
    MANGAKA = 'mangaka'
    PRODUCER = 'producer'



"""Submodel for `ban.py` and model for `/api/comments`."""
from datetime import datetime

from pydantic import BaseModel

from .user import UserInfo


class CommentInfo(BaseModel):
    """Represents a comment info entity.

    Used for `ban.py` model
    """
    id: int
    user_id: int
    commentable_id: int
    commentable_type: str
    body: str
    created_at: datetime
    updated_at: datetime
    is_offtopic: bool


class Comment(CommentInfo):
    """Represents a comment entity.

    Used for `/api/comments`
    """
    html_body: str
    can_be_edited: bool
    user: UserInfo



"""Exception for raising on too many requests."""

from loguru import logger


class RetryLater(Exception):

    def __init__(self, error_message: str):
        logger.warning(error_message)
        super().__init__(error_message)



"""Submodel for `stats.py`."""
from typing import List, Optional

from pydantic import BaseModel

from .rating import Rating


class RatingList(BaseModel):
    """Represents ratings collection of anime/manga."""
    anime: Optional[List[Rating]]
    manga: Optional[List[Rating]]



"""Represents `/api/characters` resource."""
from typing import Any, Dict, List, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..exceptions import ShikimoriAPIResponseError
from ..models import CharacterInfo, Character
from ..utils import Utils
from .base_resource import BaseResource


class Characters(BaseResource):
    """Characters resource class.

    Used to represent `/api/characters` resource
    """

    @method_endpoint('/api/characters/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, character_id: int):
        """Returns character info by ID.

        :param character_id: ID of character to get info
        :type character_id: int

        :return: Character info
        :rtype: Optional[Character]
        """
        response = await self._client.request(
            self._client.endpoints.character(character_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Character)

    @method_endpoint('/api/characters/search')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def search(self, search: Optional[str] = None):
        """Returns list of found characters.

        :param search: Search query for characters
        :type search: Optional[str]

        :return: List of found characters
        :rtype: List[CharacterInfo]
        """
        query_dict = Utils.create_query_dict(search=search)

        response = await self._client.request(
            self._client.endpoints.character_search, query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=CharacterInfo)



"""Shikithon API Module.

This is main module with a class
for interacting with the Shikimori API
"""
import sys
from typing import Optional, TypeVar

from loguru import logger

from .base_client import Client
from .resources import (AbuseRequests, Achievements, Animes, Appears, Bans,
                        Calendars, Characters, Clubs, Comments, Constants,
                        Dialogs, Favorites, Forums, Friends, Genres, Mangas,
                        Messages, People, Publishers, Ranobes, Reviews, Stats,
                        Studios, Styles, Topics, UserImages, UserRates, Users)
from .store import NullStore, Store

RT = TypeVar('RT')


class ShikimoriAPI(Client):
    """Main class for interacting with the API.

    Current API class uses base client for interacting with API.
    Also, all API methods splitted up to resources for convinient usage
    """

    __slots__ = ('achievements', 'animes', 'appears', 'bans', 'calendars',
                 'characters', 'clubs', 'comments', 'constants', 'dialogs',
                 'favorites', 'forums', 'friends', 'genres', 'mangas',
                 'messages', 'people', 'publishers', 'ranobes', 'reviews',
                 'stats', 'studios', 'styles', 'topics', 'user_images',
                 'user_rates', 'users', 'abuse_requests')

    def __init__(self,
                 app_name: str = 'Api Test',
                 store: Store = NullStore(),
                 auto_close_store: bool = True,
                 logging: Optional[bool] = False):
        """Shikimori API class initialization.

        This magic method inits client and all resources
        for interacting with.

        :param app_name: OAuth App name
        :type app_name: str

        :param store: Class instance for store configs
        :type store: Optional[Store]

        :param auto_close_store: Auto close store option
        :type auto_close_store: bool

        :param logging: Logging flag
        :type logging: Optional[bool]
        """
        if logging:
            logger.configure(handlers=[{
                'sink':
                    sys.stdout,
                'level':
                    'INFO',
                'format':
                    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> ' \
                    '| <blue>{level}</blue> | <level>{message}</level>',
                'colorize':
                    True,
            }, {
                'sink': 'shikithon_{time}.log',
                'level': 'DEBUG',
                'format': '{time} | {level} | {file}.{function}: {message}',
                'rotation': '5 MB',
                'compression': 'zip',
            }])
        if not logging:
            logger.disable('shikithon')

        logger.info('Initializing API object')

        super().__init__(app_name, store, auto_close_store)

        self.achievements = Achievements(self)
        self.animes = Animes(self)
        self.appears = Appears(self)
        self.bans = Bans(self)
        self.calendars = Calendars(self)
        self.characters = Characters(self)
        self.clubs = Clubs(self)
        self.comments = Comments(self)
        self.constants = Constants(self)
        self.dialogs = Dialogs(self)
        self.favorites = Favorites(self)
        self.forums = Forums(self)
        self.friends = Friends(self)
        self.genres = Genres(self)
        self.mangas = Mangas(self)
        self.messages = Messages(self)
        self.people = People(self)
        self.publishers = Publishers(self)
        self.ranobes = Ranobes(self)
        self.reviews = Reviews(self)
        self.stats = Stats(self)
        self.studios = Studios(self)
        self.styles = Styles(self)
        self.topics = Topics(self)
        self.user_images = UserImages(self)
        self.user_rates = UserRates(self)
        self.users = Users(self)
        self.abuse_requests = AbuseRequests(self)

        logger.info('Successfully initialized API object')



"""Submodel for `people.py`."""
from typing import Optional

from pydantic import BaseModel


class Date(BaseModel):
    """Date object model.

    Used to represent birthday or decease of person
    """
    day: Optional[int]
    month: Optional[int]
    year: Optional[int]



"""Represents `/api/mangas` resource."""
from typing import Any, Dict, List, Optional, Union, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import (MangaCensorship, MangaKind, MangaList, MangaOrder,
                     MangaStatus)
from ..exceptions import ShikimoriAPIResponseError
from ..models import FranchiseTree, Link, Manga, MangaInfo, RanobeInfo, Relation, Role, Topic
from ..utils import Utils
from .base_resource import BaseResource


class Mangas(BaseResource):
    """Mangas resource class.

    Used to represent `/api/mangas` resource
    """

    @method_endpoint('/api/mangas')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      order: Optional[MangaOrder] = None,
                      kind: Optional[Union[MangaKind, List[MangaKind]]] = None,
                      status: Optional[Union[MangaStatus,
                                             List[MangaStatus]]] = None,
                      season: Optional[Union[str, List[str]]] = None,
                      score: Optional[int] = None,
                      genre: Optional[Union[int, List[int]]] = None,
                      publisher: Optional[Union[int, List[int]]] = None,
                      franchise: Optional[Union[int, List[int]]] = None,
                      censored: Optional[MangaCensorship] = None,
                      mylist: Optional[Union[MangaList,
                                             List[MangaList]]] = None,
                      ids: Optional[Union[int, List[int]]] = None,
                      exclude_ids: Optional[Union[int, List[int]]] = None,
                      search: Optional[str] = None):
        """Returns mangas list.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param order: Type of order in list
        :type order: Optional[MangaOrder]

        :param kind: Type(s) of manga topic
        :type kind: Optional[Union[MangaKind, List[MangaKind]]

        :param status: Type(s) of manga status
        :type status: Optional[Union[MangaStatus, List[MangaStatus]]]

        :param season: Name(s) of manga seasons
        :type season: Optional[Union[str, List[str]]]

        :param score: Minimal manga score
        :type score: Optional[int]

        :param publisher: Publisher(s) ID
        :type publisher: Optional[Union[int, List[int]]

        :param genre: Genre(s) ID
        :type genre: Optional[Union[int, List[int]]

        :param franchise: Franchise(s) ID
        :type franchise: Optional[Union[int, List[int]]

        :param censored: Type of manga censorship
        :type censored: Optional[MangaCensorship]

        :param mylist: Status(-es) of manga in current user list.
            If app is in restricted mode,
            this parameter won't affect on response.
        :type mylist: Optional[Union[MangaList, List[MangaList]]]

        :param ids: Manga(s) ID to include
        :type ids: Optional[Union[int, List[int]]

        :param exclude_ids: Manga(s) ID to exclude
        :type exclude_ids: Optional[Union[int, List[int]]

        :param search: Search phrase to filter mangas by name
        :type search: Optional[str]

        :return: List of Mangas
        :rtype: List[MangaInfo]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             order=order,
                                             kind=kind,
                                             status=status,
                                             season=season,
                                             score=score,
                                             genre=genre,
                                             publisher=publisher,
                                             franchise=franchise,
                                             censored=censored,
                                             mylist=mylist,
                                             ids=ids,
                                             exclude_ids=exclude_ids,
                                             search=search)

        response = await self._client.request(self._client.endpoints.mangas,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=MangaInfo)

    @method_endpoint('/api/mangas/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, manga_id: int):
        """Returns info about certain manga.

        :param manga_id: Manga ID to get info
        :type manga_id: int

        :return: Manga info
        :rtype: Optional[Manga]
        """
        response = await self._client.request(
            self._client.endpoints.manga(manga_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Manga)

    @method_endpoint('/api/mangas/:id/roles')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def roles(self, manga_id: int):
        """Returns roles info of certain manga.

        :param manga_id: Manga ID to get roles
        :type manga_id: int

        :return: List of manga roles
        :rtype: List[Role]
        """
        response = await self._client.request(
            self._client.endpoints.manga_roles(manga_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Role)

    @method_endpoint('/api/mangas/:id/similar')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def similar(self, manga_id: int):
        """Returns list of similar mangas or ranobe for certain manga.

        :param manga_id: Manga ID to get similar mangas/ranobe
        :type manga_id: int

        :return: List of similar mangas/ranobe
        :rtype: List[Union[MangaInfo, RanobeInfo]]
        """
        response = await self._client.request(
            self._client.endpoints.similar_mangas(manga_id))

        return Utils.parse_mixed_response(response, List[Union[MangaInfo,
                                                               RanobeInfo]])

    @method_endpoint('/api/mangas/:id/related')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def related(self, manga_id: int):
        """Returns list of related content of certain manga.

        :param manga_id: Manga ID to get related content
        :type manga_id: int

        :return: List of relations
        :rtype: List[Relation]
        """
        response = await self._client.request(
            self._client.endpoints.manga_related_content(manga_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Relation)

    @method_endpoint('/api/mangas/:id/franchise')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def franchise(self, manga_id: int):
        """Returns franchise tree of certain manga.

        :param manga_id: Manga ID to get franchise tree
        :type manga_id: int

        :return: Franchise tree of certain manga
        :rtype: Optional[FranchiseTree]
        """
        response = await self._client.request(
            self._client.endpoints.manga_franchise_tree(manga_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=FranchiseTree)

    @method_endpoint('/api/mangas/:id/external_links')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def external_links(self, manga_id: int):
        """Returns list of external links of certain manga.

        :param manga_id: Manga ID to get external links
        :type manga_id: int

        :return: List of external links
        :rtype: List[Link]
        """
        response = await self._client.request(
            self._client.endpoints.manga_external_links(manga_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Link)

    @method_endpoint('/api/mangas/:id/topics')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def topics(self,
                     manga_id: int,
                     page: Optional[int] = None,
                     limit: Optional[int] = None):
        """Returns list of topics of certain manga.

        :param manga_id: Manga ID to get topics
        :type manga_id: int

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of topics
        :rtype: List[Topic]
        """
        query_dict = Utils.create_query_dict(page=page, limit=limit)

        response = await self._client.request(
            self._client.endpoints.manga_topics(manga_id), query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)



"""Represents `/api/dialogs` resource."""
from typing import Any, Dict, List, Union, cast

from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType
from ..exceptions import ShikimoriAPIResponseError
from ..models import Dialog, Message
from ..utils import Utils
from .base_resource import BaseResource


class Dialogs(BaseResource):
    """Dialogs resource class.

    Used to represent `/api/dialogs` resource
    """

    @method_endpoint('/api/dialogs')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self):
        """Returns list of current user's dialogs.

        :return: List of dialogs
        :rtype: List[Dialog]
        """
        response = await self._client.request(self._client.endpoints.dialogs)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Dialog)

    @method_endpoint('/api/dialogs/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get(self, user_id: Union[int, str]):
        """Returns list messages with certain user.

        :param user_id: ID/Nickname of the user to get dialog
        :type user_id: Union[int, str]

        :return: List of messages
        :rtype: List[Message]
        """
        response = await self._client.request(
            self._client.endpoints.dialog(user_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Message)

    @method_endpoint('/api/dialogs/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, user_id: Union[int, str]):
        """Deletes dialog with certain user.

        Instead of returning just response code,
        API method returns dictionary with "notice"
        field, so it's being logged out with INFO level

        :param user_id: ID/Nickname of the user to delete dialog
        :type user_id: Union[int, str]

        :return: Status of message deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.dialog(user_id),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True



"""Model for `/api/users/:id/history`."""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .manga import MangaInfo
from .ranobe import RanobeInfo


class History(BaseModel):
    """Represents user history timeline entity."""
    id: int
    created_at: datetime
    description: str
    target: Optional[Union[AnimeInfo, MangaInfo, RanobeInfo]]



"""Shikimori API Endpoints.

This module contains Endpoints class, which
contains all endpoints for API and can form
customized endpoints via input parameters
"""
from typing import Optional, Union

from .utils import Utils


class Endpoints:
    """Contains endpoints for Shikimori API.

    This class allows to form customized endpoint depending
    on input parameters
    """

    def __init__(self, base_url: str, base_url_v2: str, oauth_url: str):
        """Initializing URLs for Shikimori's API/OAuth.

        This constructor also has `base_url_v2`, which is
        modified `base_url` that routes to new API methods

        :param base_url: URL for Shikimori API
        :type base_url: str

        :param base_url_v2: URL for Shikimori API (v.2)
        :type base_url_v2: str

        :param oauth_url: URL for Shikimori OAuth
        :type oauth_url: str
        """
        self._base_url = base_url
        self._base_url_v2 = base_url_v2
        self._oauth_url = oauth_url

    @property
    def base_url(self) -> str:
        """Returns `base_url`.

        :return: Link for Shikimori API
        :rtype: str
        """
        return self._base_url

    @base_url.setter
    def base_url(self, base_url: str):
        """Sets `base_url`.

        :param base_url: Link for Shikimori API
        :type base_url: str
        """
        self._base_url = base_url

    @property
    def base_url_v2(self) -> str:
        """Returns `base_url_v2`.

        :return: Link for Shikimori API (v.2)
        :rtype: str
        """
        return self._base_url_v2

    @base_url_v2.setter
    def base_url_v2(self, base_url_v2: str):
        """Sets `base_url_v2`.

        :param base_url_v2: Link for Shikimori API (v.2)
        :type base_url_v2: str
        """
        self._base_url_v2 = base_url_v2

    @property
    def oauth_url(self) -> str:
        """Returns `oauth_url`.

        :return: Link for Shikimori OAuth
        :rtype: str
        """
        return self._oauth_url

    @oauth_url.setter
    def oauth_url(self, oauth_url: str):
        """Sets `oauth_url`.

        :param oauth_url: Link for Shikimori OAuth
        :type oauth_url: str
        """
        self._oauth_url = oauth_url

    @property
    def oauth_token(self) -> str:
        """Returns endpoint for OAuth token.

        :return: Link for Shikimori OAuth token endpoint
        :rtype: str
        """
        return f'{self.oauth_url}/token'

    @property
    def oauth_authorize(self) -> str:
        """Returns endpoint for OAuth authorization.

        :return: Link for Shikimori OAuth authorization endpoint
        :rtype: str
        """
        return f'{self.oauth_url}/authorize'

    def authorization_link(self, client_id: str, redirect_uri: str,
                           scopes: str) -> str:
        """Returns link for getting authorization code.

        Since you need to pass a captcha to log in to Shikimori,
        there is no automatic authorization method.
        So, the user needs to pass the authorization code manually

        :param client_id: Client ID of an OAuth App
        :type client_id: str

        :param redirect_uri: Redirect URI of an OAuth App
        :type redirect_uri: str

        :param scopes: Scopes of an OAuth App
        :type scopes: str

        :return: Link for getting authorization code
        :rtype: str
        """
        query_str = Utils.convert_to_query_string({
            'client_id': client_id,
            'redirect_uri': redirect_uri,
            'response_type': 'code',
            'scope': scopes
        })
        return f'{self.oauth_authorize}{query_str}'

    @property
    def abuse_requests(self) -> str:
        """Returns endpoint for abuse requests methods.

        :return: Abuse request main endpoint link
        :rtype: str
        """
        return f'{self.base_url_v2}/abuse_requests'

    @property
    def abuse_offtopic(self) -> str:
        """Returns endpoint for marking comment as offtopic.

        :return: Abuse offtopic endpoint link
        :rtype: str
        """
        return f'{self.abuse_requests}/offtopic'

    @property
    def abuse_review(self) -> str:
        """Returns endpoint for convering comment to review.

        :return: Abuse review endpoint link
        :rtype: str
        """
        return f'{self.abuse_requests}/review'

    @property
    def abuse_violation(self) -> str:
        """Returns endpoint for creating request about violation of site rules.

        :return: Abuse violation endpoint link
        :rtype: str
        """
        return f'{self.abuse_requests}/abuse'

    @property
    def abuse_spoiler(self) -> str:
        """Returns endpoint for creating request about spoiler in content.

        :return: Abuse spoiler endpoint link
        :rtype: str
        """
        return f'{self.abuse_requests}/spoiler'

    @property
    def achievements(self) -> str:
        """Returns endpoint of user achievements.

        :return: Achievements endpoint link
        :rtype: str
        """
        return f'{self.base_url}/achievements'

    @property
    def animes(self) -> str:
        """Returns endpoint of the animes list.

        :return: Animes list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/animes'

    def anime(self, anime_id: int) -> str:
        """Returns endpoint of a specific anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime endpoint link
        :rtype: str
        """
        return f'{self.animes}/{anime_id}'

    def anime_roles(self, anime_id: int) -> str:
        """Returns endpoint of a list of roles of a specific anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime roles endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/roles'

    def similar_animes(self, anime_id: int) -> str:
        """Returns endpoint of a list of similar anime of a certain anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Similar animes endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/similar'

    def anime_related_content(self, anime_id: int) -> str:
        """Returns endpoint of a list of similar anime of a certain anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime related content endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/related'

    def anime_screenshots(self, anime_id: int) -> str:
        """Returns endpoint of screenshots of a specific anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime screenshots endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/screenshots'

    def anime_franchise_tree(self, anime_id: int) -> str:
        """Returns endpoint of franchise tree of a certain anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime franchise tree endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/franchise'

    def anime_external_links(self, anime_id: int) -> str:
        """Returns endpoint of a list of external links of a certain anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime external links endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/external_links'

    def anime_topics(self, anime_id: int) -> str:
        """Returns endpoint of a list of topics of a certain anime.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime topics endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/topics'

    def anime_videos(self, anime_id: int) -> str:
        """Returns endpoint of the anime's videos list.

        This endpoint also used for creating videos
        for anime entity.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :return: Anime's videos list endpoint link
        :rtype: str
        """
        return f'{self.anime(anime_id)}/videos'

    def anime_video(self, anime_id: int, video_id: int) -> str:
        """Returns endpoint for destroying anime's video.

        :param anime_id: Anime ID for endpoint
        :type anime_id: int

        :param video_id: Video ID for endpoint
        :type video_id: int

        :return: Anime's video destroy endpoint link
        :rtype: str
        """
        return f'{self.anime_videos(anime_id)}/{video_id}'

    @property
    def appears(self) -> str:
        """Returns endpoint for marking comments/topics as read.

        :return: Appears endpoint link
        :rtype: str
        """
        return f'{self.base_url}/appears'

    @property
    def bans_list(self) -> str:
        """Returns endpoint of the bans list.

        :return: Bans list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/bans'

    @property
    def calendar(self) -> str:
        """Returns endpoint for events calendar.

        :return: Calendar endpoint link
        :rtype: str
        """
        return f'{self.base_url}/calendar'

    def character(self, character_id: int) -> str:
        """Returns endpoint for a certain character.

        :param character_id: Character ID for endpoint
        :type character_id: int

        :return: Character endpoint link
        :rtype: str
        """
        return f'{self.base_url}/characters/{character_id}'

    @property
    def character_search(self) -> str:
        """Returns endpoint for characters search.

        :return: Characters search endpoint link
        :rtype: str
        """
        return f'{self.base_url}/characters/search'

    @property
    def clubs(self) -> str:
        """Returns endpoint of the clubs list.

        :return: Clubs list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/clubs'

    def club(self, club_id: int) -> str:
        """Returns endpoint of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club endpoint link
        :rtype: str
        """
        return f'{self.clubs}/{club_id}'

    def club_animes(self, club_id: int) -> str:
        """Returns endpoint of a list of animes of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club animes endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/animes'

    def club_mangas(self, club_id: int) -> str:
        """Returns endpoint of a list of mangas of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club mangas endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/mangas'

    def club_ranobe(self, club_id: int) -> str:
        """Returns endpoint of a list of ranobe of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club ranobe endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/ranobe'

    def club_characters(self, club_id: int) -> str:
        """Returns endpoint of a list of characters of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club characters endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/characters'

    def club_collections(self, club_id: int) -> str:
        """Returns endpoint of a list of collections of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club collections endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/collections'

    def club_clubs(self, club_id: int) -> str:
        """Returns endpoint of a list of clubs of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club clubs endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/clubs'

    def club_members(self, club_id: int) -> str:
        """Returns endpoint of a list of members of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club members endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/members'

    def club_images(self, club_id: int) -> str:
        """Returns endpoint of a list of images of a certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club images endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/images'

    def club_join(self, club_id: int) -> str:
        """Returns endpoint for joining the certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club join endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/join'

    def club_leave(self, club_id: int) -> str:
        """Returns endpoint for leaving the certain club.

        :param club_id: Club ID for endpoint
        :type club_id: int

        :return: Club leave endpoint link
        :rtype: str
        """
        return f'{self.club(club_id)}/leave'

    @property
    def comments(self) -> str:
        """Returns endpoint of the comments list.

        :return: Comments list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/comments'

    def comment(self, comment_id: int) -> str:
        """Returns endpoint of a certain comment.

        :param comment_id: Comment ID for endpoint
        :type comment_id: int

        :return: Comment endpoint link
        :rtype: str
        """
        return f'{self.comments}/{comment_id}'

    @property
    def constants(self) -> str:
        """Returns endpoint of the certain API constants.

        :return: API constants endpoint link
        :rtype: str
        """
        return f'{self.base_url}/constants'

    @property
    def anime_constants(self) -> str:
        """Returns endpoint of the anime constants.

        :return: Anime constants endpoint link
        :rtype: str
        """
        return f'{self.constants}/anime'

    @property
    def manga_constants(self) -> str:
        """Returns endpoint of the manga constants.

        :return: Manga constants endpoint link
        :rtype: str
        """
        return f'{self.constants}/manga'

    @property
    def user_rate_constants(self) -> str:
        """Returns endpoint of the user rate constants.

        :return: User rate constants endpoint link
        :rtype: str
        """
        return f'{self.constants}/user_rate'

    @property
    def club_constants(self) -> str:
        """Returns endpoint of the club constants.

        :return: Club constants endpoint link
        :rtype: str
        """
        return f'{self.constants}/club'

    @property
    def smileys_constants(self) -> str:
        """Returns endpoint of the smileys constants.

        :return: Smileys constants endpoint link
        :rtype: str
        """
        return f'{self.constants}/smileys'

    @property
    def dialogs(self) -> str:
        """Returns endpoint of the dialogs list.

        :return: Dialogs list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/dialogs'

    def dialog(self, dialog_id: Union[int, str]) -> str:
        """Returns endpoint of a certain dialog with user.

        :param dialog_id: User ID/Nickname for endpoint
        :type dialog_id: Union[int, str]

        :return: Dialog endpoint link
        :rtype: str
        """
        return f'{self.dialogs}/{dialog_id}'

    @property
    def episode_notifications(self) -> str:
        """Returns endpoint for notifying Shikimori about anime episode release.

        To use this endpoint, you need special private token,
        required to access this API

        :return: Episode notifications endpoint link
        :rtype: str
        """
        return f'{self.base_url_v2}/episode_notifications'

    def favorites_create(self, linked_type: str, linked_id: int,
                         kind: Optional[str]) -> str:
        """Returns endpoint for creating some type
        of object as favorite.

        :param linked_type: Type of object
        :type linked_type: str

        :param linked_id: ID of linked object
        :type linked_id: int

        :param kind: Kind of linked object
        (Required when linked_type is `Person`)
        :type kind: Optional[str]

        :return: Favorite create endpoint link
        :rtype: str
        """
        return f'{self.base_url}/favorites/' \
               f'{linked_type}/{linked_id}' \
               f'{"" if not kind else f"/{kind}"}'

    def favorites_destroy(self, linked_type: str, linked_id: int) -> str:
        """Returns endpoint for destroying some type
        of object from favorites.

        :param linked_type: Type of object
        :type linked_type: str

        :param linked_id: ID of linked object
        :type linked_id: int

        :return: Favorite destroy endpoint link
        :rtype: str
        """
        return f'{self.base_url}/favorites/' \
               f'{linked_type}/{linked_id}'

    def favorites_reorder(self, favorite_id: int) -> str:
        """Returns endpoint for assigning new position of a favorite.

        :param favorite_id: Favorite ID for endpoint
        :type favorite_id: int

        :return: Favorite reorder endpoint link
        :rtype: str
        """
        return f'{self.base_url}/favorites/{favorite_id}/reorder'

    @property
    def forums(self) -> str:
        """Returns endpoint of the forums list.

        :return: Forums list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/forums'

    def friend(self, friend_id: int) -> str:
        """Returns endpoint for adding/deleting friend.

        :param friend_id: Friend ID for endpoint
        :type friend_id: int

        :return: Friend addition/deletion endpoint link
        :rtype: str
        """
        return f'{self.base_url}/friends/{friend_id}'

    @property
    def genres(self) -> str:
        """Returns endpoint of the genres list.

        :return: Genres list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/genres'

    @property
    def mangas(self) -> str:
        """Returns endpoint of the mangas list.

        :return: Mangas list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/mangas'

    def manga(self, manga_id: int) -> str:
        """Returns endpoint of a specific manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga endpoint link
        :rtype: str
        """
        return f'{self.mangas}/{manga_id}'

    def manga_roles(self, manga_id: int) -> str:
        """Returns endpoint of a list of roles of a specific manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga roles endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/roles'

    def similar_mangas(self, manga_id: int) -> str:
        """Returns endpoint of a list of similar manga of a certain manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Similar mangas endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/similar'

    def manga_related_content(self, manga_id: int) -> str:
        """Returns endpoint of a list of similar manga of a certain manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga related content endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/related'

    def manga_screenshots(self, manga_id: int) -> str:
        """Returns endpoint of screenshots of a specific manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga screenshots endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/screenshots'

    def manga_franchise_tree(self, manga_id: int) -> str:
        """Returns endpoint of franchise tree of a certain manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga franchise tree endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/franchise'

    def manga_external_links(self, manga_id: int) -> str:
        """Returns endpoint of a list of external links of a certain manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga external links endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/external_links'

    def manga_topics(self, manga_id: int) -> str:
        """Returns endpoint of a list of topics of a certain manga.

        :param manga_id: Manga ID for endpoint
        :type manga_id: int

        :return: Manga topics endpoint link
        :rtype: str
        """
        return f'{self.manga(manga_id)}/topics'

    @property
    def messages(self) -> str:
        """Returns endpoint of the messages list.

        :return: Messages list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/messages'

    def message(self, message_id: int) -> str:
        """Returns endpoint of a specific message.

        :param message_id: Message ID for endpoint
        :type message_id: int

        :return: Message endpoint link
        :rtype: str
        """
        return f'{self.messages}/{message_id}'

    @property
    def messages_mark_read(self) -> str:
        """Returns endpoint for marking messages as read/unread.

        :return: Marking messages endpoint link
        :rtype: str
        """
        return f'{self.messages}/mark_read'

    @property
    def messages_read_all(self) -> str:
        """Returns endpoint for marking all messages as read.

        :return: Reading messages endpoint link
        :rtype: str
        """
        return f'{self.messages}/read_all'

    @property
    def messages_delete_all(self) -> str:
        """Returns endpoint for deleting all messages.

        :return: Deleting messages endpoint link
        :rtype: str
        """
        return f'{self.messages}/delete_all'

    def people(self, person_id: int) -> str:
        """Returns endpoint for a certain person.

        :param person_id: Person ID for endpoint
        :type person_id: int

        :return: Person endpoint link
        :rtype: str
        """
        return f'{self.base_url}/people/{person_id}'

    @property
    def people_search(self) -> str:
        """Returns endpoint for people search.

        :return: People search endpoint link
        :rtype: str
        """
        return f'{self.base_url}/people/search'

    @property
    def publishers(self) -> str:
        """Returns endpoint of the publishers list.

        :return: Publishers list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/publishers'

    @property
    def ranobes(self) -> str:
        """Returns endpoint of the ranobes list.

        :return: Ranobes list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/ranobe'

    def ranobe(self, ranobe_id: int) -> str:
        """Returns endpoint of a specific ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe endpoint link
        :rtype: str
        """
        return f'{self.ranobes}/{ranobe_id}'

    def ranobe_roles(self, ranobe_id: int) -> str:
        """Returns endpoint of a list of roles of a specific ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe roles endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/roles'

    def similar_ranobes(self, ranobe_id: int) -> str:
        """Returns endpoint of a list of similar ranobe of a certain ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Similar ranobes endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/similar'

    def ranobe_related_content(self, ranobe_id: int) -> str:
        """Returns endpoint of a list of similar ranobe of a certain ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe related content endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/related'

    def ranobe_franchise_tree(self, ranobe_id: int) -> str:
        """Returns endpoint of franchise tree of a certain ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe franchise tree endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/franchise'

    def ranobe_external_links(self, ranobe_id: int) -> str:
        """Returns endpoint of a list of external links of a certain ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe external links endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/external_links'

    def ranobe_topics(self, ranobe_id: int) -> str:
        """Returns endpoint of a list of topics of a certain ranobe.

        :param ranobe_id: Ranobe ID for endpoint
        :type ranobe_id: int

        :return: Ranobe topics endpoint link
        :rtype: str
        """
        return f'{self.ranobe(ranobe_id)}/topics'

    @property
    def reviews(self) -> str:
        """Returns endpoint of the reviews.

        :return: Reviews endpoint link
        :rtype: str
        """
        return f'{self.base_url}/reviews'

    def review(self, review_id: int) -> str:
        """Returns endpoint of a specific review.

        :param review_id: Review ID for endpoint
        :type review_id: int

        :return: Review endpoint link
        :rtype: str
        """
        return f'{self.reviews}/{review_id}'

    @property
    def active_users(self) -> str:
        """Returns endpoint of a list of users,
        having at least 1 completed animes and active during last month.

        :return: Active users list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/stats/active_users'

    @property
    def studios(self) -> str:
        """Returns endpoint of the studios list.

        :return: Studios list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/studios'

    @property
    def styles(self) -> str:
        """Returns endpoint of the styles.

        This endpoint also used for styles creation

        :return: Style create endpoint link
        :rtype: str
        """
        return f'{self.base_url}/styles'

    @property
    def style_preview(self) -> str:
        """Returns endpoint for previwing style.

        :return: Style preview endpoint link
        :rtype: str
        """
        return f'{self.styles}/preview'

    def style(self, style_id: int) -> str:
        """Returns endpoint of the style.

        :param style_id: Style ID for endpoint
        :type style_id: int

        :return: Style endpoint link
        :rtype: str
        """
        return f'{self.styles}/{style_id}'

    @property
    def topics(self) -> str:
        """Returns endpoint of the topics list.

        :return: Topics list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/topics'

    def topic(self, topic_id: int) -> str:
        """Returns endpoint of the topic.

        :param topic_id: Topic ID for endpoint
        :type topic_id: int

        :return: Topic endpoint link
        :rtype: str
        """
        return f'{self.topics}/{topic_id}'

    @property
    def updates_topics(self) -> str:
        """Returns endpoint of the topics updates list.

        :return: Topics updates list endpoint link
        :rtype: str
        """
        return f'{self.topics}/updates'

    @property
    def hot_topics(self) -> str:
        """Returns endpoint of the hot topics list.

        :return: Hot topics list endpoint link
        :rtype: str
        """
        return f'{self.topics}/hot'

    def topic_ignore(self, topic_id: int) -> str:
        """Returns endpoint for ignoring/unignoring topic.

        :param topic_id: Topic ID for endpoint
        :type topic_id: int

        :return: Ignore/unignore topic endpoint link
        :rtype: str
        """
        return f'{self.base_url_v2}/topics/{topic_id}/ignore'

    @property
    def user_images(self) -> str:
        """Returns endpoint for creating an user image.

        :return: User image create endpoint link
        :rtype: str
        """
        return f'{self.base_url}/user_images'

    @property
    def users(self) -> str:
        """Returns endpoint of the users list.

        :return: Users list endpoint link
        :rtype: str
        """
        return f'{self.base_url}/users'

    @property
    def whoami(self) -> str:
        """Returns endpoint of the current user's brief info.

        :return: Current user's brief info endpoint link
        :rtype: str
        """
        return f'{self.users}/whoami'

    @property
    def sign_out(self) -> str:
        """Returns endpoint of a user's sign out.

        :return: User's sign out endpoint link
        :rtype: str
        """
        return f'{self.users}/sign_out'

    def user(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User endpoint link
        :rtype: str
        """
        return f'{self.users}/{user_id}'

    def user_info(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's brief info.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's brief info endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/info'

    def user_friends(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's friends list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's friends list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/friends'

    def user_clubs(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's clubs list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's clubs list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/clubs'

    def user_anime_rates(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's anime rates list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's anime rates list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/anime_rates'

    def user_manga_rates(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's manga rates list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's manga rates list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/manga_rates'

    def user_favourites(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's favourites list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's favorites list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/favourites'

    def user_messages(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's messages list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's messages list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/messages'

    def user_unread_messages(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's unread messages counts.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's unread messages counts endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/unread_messages'

    def user_history(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's history list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's history list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/history'

    def user_bans(self, user_id: Union[int, str]) -> str:
        """Returns endpoint of the user's bans list.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: User's bans list endpoint link
        :rtype: str
        """
        return f'{self.user(user_id)}/bans'

    def user_ignore(self, user_id: Union[int, str]) -> str:
        """Returns endpoint for ignoring/unignoring user.

        :param user_id: User ID/Nickname for endpoint
        :type user_id: Union[int, str]

        :return: Ignore/unignore user endpoint link
        :rtype: str
        """
        return f'{self.base_url_v2}/users/{user_id}/ignore'

    @property
    def user_rates(self) -> str:
        """Returns endpoint of the user rates list.

        :return: User rates list endpoint link
        :rtype: str
        """
        return f'{self.base_url_v2}/user_rates'

    def user_rate(self, user_rate_id: int) -> str:
        """Returns endpoint of the user rate.

        :param user_rate_id: User rate ID for endpoint
        :type user_rate_id: int

        :return: User rate endpoint link
        :rtype: str
        """
        return f'{self.user_rates}/{user_rate_id}'

    def user_rate_increment(self, user_rate_id: int) -> str:
        """Returns endpoint for incrementing episodes/chapters by 1.

        :param user_rate_id: User rate ID for increment endpoint
        :type user_rate_id: int

        :return: User rate increment endpoint link
        :rtype: str
        """
        return f'{self.user_rate(user_rate_id)}/increment'

    def user_rates_cleanup(self, user_rate_type: str) -> str:
        """Returns endpoint for cleanup user rates by type.

        This endpoint is used by API v1

        :param user_rate_type: Type of an user rate
        :type user_rate_type: str

        :return: User rates cleanup endpoint link
        :rtype: str
        """
        return f'{self.base_url}/user_rates/{user_rate_type}/cleanup'

    def user_rates_reset(self, user_rate_type: str) -> str:
        """Returns endpoint for resetting user rates by type.

        This endpoint is used by API v1

        :param user_rate_type: Type of an user rate
        :type user_rate_type: str

        :return: User rates reset endpoint link
        :rtype: str
        """
        return f'{self.base_url}/user_rates/{user_rate_type}/reset'



"""Set of utility methods for the shikithon library.

This file contains the Utils class
with all the necessary utility methods
to work with the library
"""

import imghdr
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, overload

from aiohttp import ClientResponse, ClientSession, FormData
from loguru import logger
from pydantic import BaseModel, parse_obj_as
from validators import url

from ..enums import ResponseCode

LOWER_LIMIT_NUMBER = 1
M = TypeVar('M', bound=BaseModel)

R = TypeVar('R')
T = TypeVar('T')


class Utils:
    """Utils class.

    Contains all the necessary utility methods
    to work with the library
    """

    @staticmethod
    def convert_to_query_string(query_dict: Optional[Dict[str, str]]):
        """Converts query dict to query string for endpoint link.

        If query_dict is None or empty, returns empty string

        :param query_dict: Query dictionary
        :type query_dict: Dict[str, str]

        :return: Query string
        :rtype: str
        """
        logger.debug(f'Converting {query_dict} to query string')

        if not query_dict:
            logger.debug('Query dictionary is None or empty. '
                         'Returning empty string')
            return ''

        query_dict_str = '&'.join(
            f'{key}={val}' for (key, val) in query_dict.items())

        query_string = f'?{query_dict_str}'

        logger.debug(f'Formed string: "{query_string}"')
        return query_string

    @staticmethod
    async def get_image_data(image_path: Optional[str]):
        """Extracts image data from image path.

        If image_path is a link, fetch the image data from the link

        :param image_path: Path to image
        :type image_path: str

        :return: Image data
        :rtype: Optional[bytes]
        """
        logger.debug(f'Getting image from "{image_path}"')

        if image_path is None:
            logger.debug('Image path is None or empty')
            return None

        if not isinstance(url(image_path), bool):
            logger.debug('Image path is local. Reading image data')
            return Utils._extract_local_image_data(image_path)

        logger.debug('Image path is a link. Querying image data')
        return await Utils._extract_remote_image_data(image_path)

    @staticmethod
    def _extract_local_image_data(image_path: str):
        """Extracts image data from local path.

        :param image_path: Path to local image
        :type image_path: str

        :return: Image data
        :rtype: Optional[bytes]
        """
        try:
            image_path_test = imghdr.what(image_path)
            if image_path_test is None:
                logger.warning('Passed image path is not a valid image')
                return None
            with open(image_path, 'rb') as image_file:
                image_data = image_file.read()
            logger.debug('Successfully extracted image data')
            return image_data
        except FileNotFoundError:
            logger.warning('Passed image file is not exists')
            return None

    @staticmethod
    async def _extract_remote_image_data(image_url: str):
        """Extracts image data from remote path.

        :param image_url: Path to remote image
        :type image_url: str

        :return: Image data
        :rtype: Optional[bytes]
        """
        async with ClientSession() as session:
            image_response = await session.get(image_url)
            if not image_response.content_type.startswith('image/'):
                logger.warning('Passed URL is not a valid image URL')
                return None
            logger.debug('Successfully extracted image data')
            return await image_response.read()

    @staticmethod
    def create_query_dict(**params_data: Optional[Any]):
        """Creates query dictionary for API request.

        This methods checks for data types and converts to valid one

        :param params_data: Parameters data for API request
        :type params_data: Optional[Any]

        :return: Query dictionary
        :rtype: Dict[str, str]
        """
        logger.debug(
            'Generating query dictionary for request. ' \
            f'Passed {params_data=}'
        )

        query_dict: Dict[str, str] = {}

        logger.debug('Extracting data for query dictionary')
        for key, data in params_data.items():
            if data is None:
                continue
            query_dict.update({key: Utils._convert_dictionary_value(data)})

        logger.debug(f'Generated query dictionary: {query_dict=}')
        return query_dict

    @staticmethod
    def create_data_dict(**dict_data: Optional[Any]):
        """Creates data dictionary for API request.

        This methods checks for data types and converts to valid one

        If dict_data doesn't contain "dict_name" key,
        generated dictionary will have "temp" as root dictionary name,
        which will be removed on return

        :param dict_data: Body data for API request
        :type dict_data: Optional[Any]

        :return: Data dictionary
        :rtype: Union[Dict[str, Dict[str, str]], Dict[str, str]]
        """
        logger.debug(
            f'Generating data dictionary for request. Passed {dict_data=}')

        logger.debug('Extracting root dictionary name')
        data_dict_name: Optional[str] = dict_data.pop('dict_name', None)

        if data_dict_name is None:
            logger.debug(
                'There is no dict_name in dict_data. ' \
                'Returning dictionary without root'
            )
            return Utils.create_query_dict(**dict_data)

        logger.debug(f'Setting root dictionary with name "{data_dict_name}"')
        new_data_dict: Dict[str, Dict[str, str]] = {data_dict_name: {}}

        logger.debug('Extracting data for data dictionary')
        for key, data in dict_data.items():
            if data is None:
                continue
            new_data_dict[data_dict_name].update(
                {key: Utils._convert_dictionary_value(data, data_dict=True)})

        logger.debug(f'Generated data dictionary: {new_data_dict}')
        return new_data_dict

    @staticmethod
    def _convert_dictionary_value(dict_value: Any, data_dict: bool = False):
        """Converts dictionary value to string.

        If data_dict is False, converts list values to comma-separated string.
        Otherwise, returns list value as is

        :param dict_value: Dictionary value
        :type dict_value: Any

        :param data_dict: Flag if checking value for data dictionary
        :type data_dict: bool

        :return: Converted value
        :rtype: str
        """
        logger.debug(f'Converting value "{dict_value}" to string')

        if isinstance(dict_value, bool):
            return str(int(dict_value))
        elif isinstance(dict_value, int):
            return str(dict_value)
        elif isinstance(dict_value, (list, tuple)):
            if data_dict is True:
                return dict_value
            return ','.join([str(x) for x in dict_value])
        elif isinstance(dict_value, bytes):
            return dict_value
        else:
            # If something else passed, trying to convert to string
            return str(dict_value)

    @staticmethod
    def validate_response_code(response_code: Union[int, Any],
                               check_code: ResponseCode):
        """Validates passed response code.

        :param response_code: Passed response code
        :type response_code: Union[int, Any]

        :param check_code: Response code to compare with
        :type check_code: ResponseCode

        :return: Validated response code
        :rtype: bool
        """
        logger.debug(f'Validating response code: {response_code=}, '
                     f'{check_code=}')

        if not isinstance(response_code, int):
            logger.warning(
                'Got non-int value. Cancel comparing with response code')
            return False

        response_code_value: int = check_code.value
        return response_code == response_code_value

    @overload
    @staticmethod
    def validate_response_data(
        response_data: Dict[str, Any],
        data_model: Type[M],
    ) -> Optional[M]:
        ...

    @overload
    @staticmethod
    def validate_response_data(
        response_data: List[Dict[str, Any]],
        data_model: Type[M],
    ) -> List[M]:
        ...

    @staticmethod
    def validate_response_data(
        response_data: Union[Dict[str, Any], List[Dict[str, Any]]],
        data_model: Type[M],
    ) -> Optional[Union[Optional[M], List[M]]]:
        """Validates passed response data and returns parsed models.

        :param response_data: Passed response data
        :type response_data: Union[Dict[str, Any], List[Dict[str, Any]]]

        :param data_model: Model to convert into passed response data
        :type data_model: Type[M]

        :return: Parsed response data
        :rtype: Optional[Union[List, Optional[M], List[M]]]
        """
        logger.debug('Validating and parsing response data '\
                    f'using "{data_model.__name__}" data model')
        logger.debug(f'Passed response data: {response_data}')

        if not response_data:
            logger.debug('Response data is empty. Returning')
            return [] if isinstance(response_data, list) else None

        return [data_model(**item) for item in response_data] if isinstance(
            response_data, list) else data_model(**response_data)

    @staticmethod
    def parse_mixed_response(response: Any, parse_type: Type[T]):
        """Parses response, mixed with other entities,
        that can't be parsed with validation utility.

        Due to fact, that every Manga and Ranobe can have both models as
        similar, this utility method helps parse response correctly

        :param response: Passed response data
        :type response: Any

        :param parse_type: Type for parsing response to
        :type parse_type: Type[T]

        :return: Parsed response to passed type
        :rtype: T
        """
        logger.info('Parsing response with mixed models')
        logger.info(f'Parsing using type: {parse_type}')
        return parse_obj_as(parse_type, response)

    @staticmethod
    def create_form_data(raw_data: Dict[str, Any]):
        """Creates form data for API request.

        Method converts dictionary with data to
        FormData object for API request

        :param raw_data: Raw data for API request
        :type raw_data: Dict[str, Any]

        :return: Form data
        :rtype: Optional[FormData]
        """
        if not isinstance(raw_data, dict):
            logger.debug('Passed data is not a dictionary')
            return None

        if any(isinstance(value, dict) for value in raw_data.values()):
            return Utils._generate_nested_form_data(raw_data)
        return Utils._generate_plain_form_data(raw_data)

    @staticmethod
    def _generate_nested_form_data(raw_data: Dict[str, Dict[str, Any]]):
        """Generates FormData for nested dictionaries.

        :param raw_data: Raw data for API request
        :type raw_data: Dict[str, Dict[str, Any]]

        :return: Generated FormData
        :rtype: FormData
        """
        form_data = FormData()
        logger.debug('Generating FormData for nested dictionaries')

        for key, value in raw_data.items():
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, list):
                        for list_value in subvalue:
                            form_data.add_field(f'{key}[{subkey}][]',
                                                list_value)
                    else:
                        form_data.add_field(f'{key}[{subkey}]', subvalue)
            else:
                form_data.add_field(key, value)

        logger.debug('Successfully created FormData object')
        return form_data

    @staticmethod
    def _generate_plain_form_data(raw_data: Dict[str, Any]):
        """Generates FormData for plain dictionary.

        :param raw_data: Raw data for API request
        :type raw_data: Dict[str, Any]

        :return: Generated FormData
        :rtype: FormData
        """
        form_data = FormData()
        logger.debug('Generating FormData for plain dictionary')

        for key, value in raw_data.items():
            if isinstance(value, list):
                for list_value in value:
                    form_data.add_field(f'{key}[]', list_value)
            form_data.add_field(key, value)

        logger.debug('Successfully generated FormData object')
        return form_data

    @staticmethod
    async def log_response_info(response: ClientResponse):
        """Logs response info.

        This method extracts response status, headers and data

        :param response: Response object
        :type response: ClientResponse
        """
        logger.debug(f'Response status: {response.status}')
        logger.debug(f'Response headers: {response.headers}')
        logger.debug(f'Response data: {await response.text()}')



"""Exception for raising if request response status is not lower than 400."""

from loguru import logger


class ShikimoriAPIResponseError(Exception):

    def __init__(self, method: str, status: int, url: str, text: str):
        error_message = f'{method} {status} {url}\n{text}'
        logger.error(error_message)
        super().__init__(error_message)



"""Model for `/api/animes`."""
from datetime import date, datetime
from typing import List, Optional

from pydantic import BaseModel

from .genre import Genre
from .image import Image
from .screenshot import Screenshot
from .studio import Studio
from .user_rate import UserRate
from .user_rate_score import UserRateScore
from .user_rate_status import UserRateStatus
from .video import Video


class AnimeInfo(BaseModel):
    """Represents an anime info entity."""
    id: int
    name: str
    russian: str
    image: Image
    url: str
    kind: Optional[str]
    score: float
    status: str
    episodes: int
    episodes_aired: int
    aired_on: Optional[date]
    released_on: Optional[date]


class Anime(AnimeInfo):
    """Represents an anime entity."""
    rating: str
    english: List[Optional[str]]
    japanese: List[Optional[str]]
    synonyms: List[str]
    license_name_ru: Optional[str]
    duration: int
    description: Optional[str]
    description_html: str
    description_source: Optional[str]
    franchise: Optional[str]
    favoured: bool
    anons: bool
    ongoing: bool
    thread_id: Optional[int]
    topic_id: Optional[int]
    myanimelist_id: int
    rates_scores_stats: List[UserRateScore]
    rates_statuses_stats: List[UserRateStatus]
    updated_at: datetime
    next_episode_at: Optional[datetime]
    fansubbers: List[str]
    fandubbers: List[str]
    licensors: List[str]
    genres: List[Genre]
    studios: List[Studio]
    videos: List[Video]
    screenshots: List[Screenshot]
    user_rate: Optional[UserRate]


class CharacterAnime(AnimeInfo):
    """Represents a character anime info entity."""
    roles: List[str]
    role: str



"""API resources for shikithon API class."""

from .abuse_requests import AbuseRequests
from .achievements import Achievements
from .animes import Animes
from .appears import Appears
from .bans import Bans
from .calendars import Calendars
from .characters import Characters
from .clubs import Clubs
from .comments import Comments
from .constants import Constants
from .dialogs import Dialogs
from .favorites import Favorites
from .forums import Forums
from .friends import Friends
from .genres import Genres
from .mangas import Mangas
from .messages import Messages
from .people import People
from .publishers import Publishers
from .ranobes import Ranobes
from .reviews import Reviews
from .stats import Stats
from .studios import Studios
from .styles import Styles
from .topics import Topics
from .user_images import UserImages
from .user_rates import UserRates
from .users import Users

__all__ = [
    'Achievements',
    'Animes',
    'Appears',
    'Bans',
    'Calendars',
    'Characters',
    'Clubs',
    'Comments',
    'Constants',
    'Dialogs',
    'Favorites',
    'Forums',
    'Friends',
    'Genres',
    'Mangas',
    'Messages',
    'People',
    'Publishers',
    'Ranobes',
    'Reviews',
    'Stats',
    'Studios',
    'Styles',
    'Topics',
    'UserImages',
    'UserRates',
    'Users',
    'AbuseRequests',
]



"""Decorator for handling method exceptions."""
from functools import wraps
from typing import Any, Awaitable, Callable, Type, TypeVar

from loguru import logger
from typing_extensions import ParamSpec

P = ParamSpec('P')
R = TypeVar('R')


def exceptions_handler(
    *exceptions: Type[Exception], **params: Any
) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
    """Decorator for handling method exceptions.

    :param exceptions: Tuple of exceptions
    :type exceptions: Type[Exception]

    :param params: Dictionary with other params
    :type params: Any

    :return: Decorator function
    :rtype: Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]
    """

    if not exceptions:
        exceptions = (Exception,)

    is_error_logging = params.get('logging')

    if is_error_logging is None:
        is_error_logging = False

    def exceptions_handler_wrapper(
            function: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        """Exceptions handler wrapper.

        :param function: Function to wrap
        :type function: Callable[P, Awaitable[R]]

        :return: Wrapped function
        :rtype: Callable[P, Awaitable[R]]
        """

        @wraps(function)
        async def exceptions_handler_wrapped(*args: P.args,
                                             **kwargs: P.kwargs) -> R:
            """Decorator's wrapped function for handling method exceptions.

            :param args: Positional arguments
            :type args: P.args

            :param kwargs: Keyword arguments
            :type kwargs: P.kwargs

            :return: Result of decorated function
            :rtype: R
            """
            logger.debug(f'Handling "{function.__qualname__}" exceptions')

            try:
                return await function(*args, **kwargs)
            except exceptions as e:
                if is_error_logging:
                    logger.error(e)
                return params['fallback']

        return exceptions_handler_wrapped

    return exceptions_handler_wrapper



"""Model for `/api/animes/:id/external_links`."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class Link(BaseModel):
    """Represents external link entity."""
    id: Optional[int]
    kind: str
    url: str
    source: str
    entry_id: int
    entry_type: str
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    imported_at: Optional[datetime]



"""Submodel for `people.py`."""
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .manga import MangaInfo
from .ranobe import RanobeInfo


class Works(BaseModel):
    """Represents works entity of person."""
    anime: Optional[AnimeInfo]
    manga: Optional[Union[MangaInfo, RanobeInfo]]
    role: str



"""Represents `/api/appears` resource."""
from typing import cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType, ResponseCode
from ..exceptions import ShikimoriAPIResponseError
from ..utils import Utils
from .base_resource import BaseResource


class Appears(BaseResource):
    """Appears resource class.

    Used to represent `/api/appears` resource
    """

    @method_endpoint('/api/appears')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def mark(self, *ids: str):
        """Marks comments or topics as read.

        :param ids: Tuple with IDs of comments
        or topics to mark
        :type ids: str

        :return: Status of mark
        :rtype: bool
        """
        data_dict = Utils.create_data_dict(ids=ids)

        response = await self._client.request(self._client.endpoints.appears,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            ResponseCode.SUCCESS)



"""Model for `api/v2/abuse_requests`."""
from typing import List

from pydantic import BaseModel


class AbuseResponse(BaseModel):
    """Represents abuse response entity."""
    kind: str
    value: bool
    affected_ids: List[int]



"""Represents `/api/topics` and `/api/v2/topics` resources."""
from typing import Any, Dict, List, Optional, cast

from loguru import logger

from ..decorators import exceptions_handler, method_endpoint
from ..enums import RequestType, TopicForumType, TopicLinkedType, TopicType
from ..exceptions import ShikimoriAPIResponseError
from ..models import Topic, TopicUpdate
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'topic'


class Topics(BaseResource):
    """Topics resource class.

    Used to represent `/api/topics` and `/api/v2/topics` resources
    """

    @method_endpoint('/api/topics')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      forum: Optional[TopicForumType] = None,
                      linked_id: Optional[int] = None,
                      linked_type: Optional[TopicLinkedType] = None,
                      topic_type: Optional[TopicType] = None):
        """Returns list of topics.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param forum: Type of forum
        :type forum: Optional[TopicForumType]

        :param linked_id: ID of linked topic (Used together with linked_type)
        :type linked_id: Optional[int]

        :param linked_type: Type of linked topic (Used together with linked_id)
        :type linked_type: Optional[TopicLinkedType]

        :param topic_type: Type of topic
        :type topic_type: Optional[TopicType]

        :return: List of topics
        :rtype: List[Topic]
        """
        data_dict = Utils.create_query_dict(page=page,
                                            limit=limit,
                                            forum=forum,
                                            linked_id=linked_id,
                                            linked_type=linked_type,
                                            type=topic_type)

        response = await self._client.request(self._client.endpoints.topics,
                                              query=data_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)

    @method_endpoint('/api/topics/updates')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def updates(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None):
        """Returns list of NewsTopics about database updates.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of topic updates
        :rtype: List[TopicUpdate]
        """
        query_dict = Utils.create_query_dict(page=page, limit=limit)

        response = await self._client.request(
            self._client.endpoints.updates_topics, query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=TopicUpdate)

    @method_endpoint('/api/topics/hot')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def hot(self, limit: Optional[int] = None):
        """Returns list of hot topics.

        :param limit: Number of results limit
        :type limit: Optional[int]

        :return: List of topics
        :rtype: List[Topic]
        """
        query_dict = Utils.create_query_dict(limit=limit)

        response = await self._client.request(self._client.endpoints.hot_topics,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)

    @method_endpoint('/api/topics/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, topic_id: int):
        """Returns info about topic.

        :param topic_id: ID of topic to get
        :type topic_id: int

        :return: Info about topic
        :rtype: Optional[Topic]
        """
        response = await self._client.request(
            self._client.endpoints.topic(topic_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Topic)

    @method_endpoint('/api/topics')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def create(self,
                     body: str,
                     forum_id: int,
                     title: str,
                     user_id: int,
                     linked_id: Optional[int] = None,
                     linked_type: Optional[TopicLinkedType] = None):
        """Creates topic.

        :param body: Body of topic
        :type body: str

        :param forum_id: ID of forum to post
        :type forum_id: int

        :param title: Title of topic
        :type title: str

        :param user_id: ID of topic creator
        :type user_id: int

        :param linked_id: ID of linked topic (Used together with linked_type)
        :type linked_id: Optional[int]

        :param linked_type: Type of linked topic (Used together with linked_id)
        :type linked_type: Optional[TopicLinkedType]

        :return: Created topic info
        :rtype: Optional[Topic]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           body=body,
                                           forum_id=forum_id,
                                           linked_id=linked_id,
                                           linked_type=linked_type,
                                           title=title,
                                           type=TopicType.REGULAR_TOPIC,
                                           user_id=user_id)

        response = await self._client.request(self._client.endpoints.topics,
                                              data=data_dict,
                                              request_type=RequestType.POST)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Topic)

    @method_endpoint('/api/topics/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self,
                     topic_id: int,
                     body: Optional[str] = None,
                     title: Optional[str] = None,
                     linked_id: Optional[int] = None,
                     linked_type: Optional[TopicLinkedType] = None):
        """Updates topic.

        :param topic_id: ID of topic to update
        :type topic_id: int

        :param body: Body of topic
        :type body: Optional[str]

        :param title: Title of topic
        :type title: Optional[str]

        :param linked_id: ID of linked topic (Used together with linked_type)
        :type linked_id: Optional[int]

        :param linked_type: Type of linked topic (Used together with linked_id)
        :type linked_type: Optional[TopicLinkedType]

        :return: Updated topic info
        :rtype: Optional[Topic]
        """
        data_dict = Utils.create_data_dict(dict_name=DICT_NAME,
                                           body=body,
                                           linked_id=linked_id,
                                           linked_type=linked_type,
                                           title=title)

        response = await self._client.request(
            self._client.endpoints.topic(topic_id),
            data=data_dict,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Topic)

    @method_endpoint('/api/topics/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def delete(self, topic_id: int):
        """Deletes topic.

        :param topic_id: ID of topic to delete
        :type topic_id: int

        :return: Status of topic deletion
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.topic(topic_id),
            request_type=RequestType.DELETE)

        logger.info(response)

        return True

    @method_endpoint('/api/v2/topics/:topic_id/ignore')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def ignore(self, topic_id: int):
        """Sets topic as ignored.

        :param topic_id: ID of topic to ignore
        :type topic_id: int

        :return: True if topic was ignored, False otherwise
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.topic_ignore(topic_id),
            request_type=RequestType.POST)

        return cast(Dict[str, Any], response).get('is_ignored', None) is True

    @method_endpoint('/api/v2/topics/:topic_id/ignore')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def unignore(self, topic_id: int):
        """Sets topic as unignored.

        :param topic_id: ID of topic to unignore
        :type topic_id: int

        :return: True if topic was unignored, False otherwise
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.topic_ignore(topic_id),
            request_type=RequestType.DELETE)

        return cast(Dict[str, Any], response).get('is_ignored', None) is False



"""Submodel for `stats.py`."""
from typing import List

from pydantic import BaseModel

from .type import Type


class TypeList(BaseModel):
    """Represents types collection of anime/manga."""
    anime: List[Type]
    manga: List[Type]



"""Exception for raising on already running client."""

from loguru import logger


class AlreadyRunningClient(Exception):

    def __init__(self):
        error_message = 'Client is already running'
        logger.error(error_message)
        super().__init__(error_message)



"""Model for `/api/users/:id/anime_rates|manga_rates`."""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .manga import MangaInfo
from .ranobe import RanobeInfo
from .user import UserInfo


class UserList(BaseModel):
    """Represents user list entity.

    Contains data of watched/read titles.
    """
    id: int
    score: int
    status: str
    text: Optional[str]
    episodes: Optional[int]
    chapters: Optional[int]
    volumes: Optional[int]
    text_html: str
    rewatches: int
    created_at: datetime
    updated_at: datetime
    user: UserInfo
    anime: Optional[AnimeInfo]
    manga: Optional[Union[MangaInfo, RanobeInfo]]



"""Model for `/api/animes/:id/topics`."""
from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel

from .anime import AnimeInfo
from .club import ClubInfo
from .character import CharacterInfo
from .critique import Critique
from .ranobe import RanobeInfo
from .forum import Forum
from .manga import MangaInfo
from .user import UserInfo


class Topic(BaseModel):
    """Represents topic entity.

    `linked` field can represent multiple models,
    based on `linked_type`

    - Anime: Used in `/api/animes/:id/topics` and `/api/topics/`
    - Manga: Used in `/api/mangas/:id/topics` and `/api/topics/`
    - Ranobe: Used in `/api/ranobe/:id/topics` and `/api/topics/`
    - Club: Used in `/api/topics`
    - Character: Used in `/api/topics`
    - Critique: Used in `/api/topics`
    - None: Used `/api/clubs/:id/collections` and `/api/topics/`
    """
    id: int
    topic_title: str
    body: str
    html_body: str
    html_footer: str
    created_at: datetime
    comments_count: int
    forum: Forum
    user: UserInfo
    type: str
    linked_id: Optional[int]
    linked_type: Optional[str]
    linked: Optional[Union[AnimeInfo, MangaInfo, RanobeInfo, ClubInfo,
                           CharacterInfo, Critique]]
    viewed: bool
    last_comment_viewed: Optional[bool]
    event: Optional[str]
    episode: Optional[int]


class TopicUpdate(BaseModel):
    """Represents topic update entity."""
    id: int
    linked: Union[AnimeInfo, MangaInfo, RanobeInfo]
    event: Optional[str]
    episode: Optional[int]
    created_at: datetime
    url: str



"""Represents `/api/clubs` resource."""
from typing import Any, Dict, List, Optional, cast

from ..decorators import exceptions_handler, method_endpoint
from ..enums import (CommentPolicy, ImageUploadPolicy, JoinPolicy, PagePolicy,
                     RequestType, ResponseCode, TopicPolicy)
from ..exceptions import ShikimoriAPIResponseError
from ..models import AnimeInfo, CharacterInfo, ClubInfo, Club, ClubImage, MangaInfo, RanobeInfo, UserInfo, Topic
from ..utils import Utils
from .base_resource import BaseResource

DICT_NAME = 'club'


class Clubs(BaseResource):
    """Clubs resource class.

    Used to represent `/api/clubs` resource
    """

    @method_endpoint('/api/clubs')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def get_all(self,
                      page: Optional[int] = None,
                      limit: Optional[int] = None,
                      search: Optional[str] = None):
        """Returns clubs list.

        :param page: Number of page
        :type page: Optional[int]

        :param limit: Number of results limit
        :type limit: Optional[int]

        :param search: Search phrase to filter clubs by name
        :type search: Optional[str]

        :return: Clubs list
        :rtype: List[ClubInfo]
        """
        query_dict = Utils.create_query_dict(page=page,
                                             limit=limit,
                                             search=search)

        response = await self._client.request(self._client.endpoints.clubs,
                                              query=query_dict)

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=ClubInfo)

    @method_endpoint('/api/clubs/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def get(self, club_id: int):
        """Returns info about club.

        :param club_id: Club ID to get info
        :type club_id: int

        :return: Info about club
        :rtype: Optional[Club]
        """
        response = await self._client.request(
            self._client.endpoints.club(club_id))

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Club)

    @method_endpoint('/api/clubs/:id')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=None)
    async def update(self,
                     club_id: int,
                     name: Optional[str] = None,
                     join_policy: Optional[JoinPolicy] = None,
                     description: Optional[str] = None,
                     display_images: Optional[bool] = None,
                     comment_policy: Optional[CommentPolicy] = None,
                     topic_policy: Optional[TopicPolicy] = None,
                     page_policy: Optional[PagePolicy] = None,
                     image_upload_policy: Optional[ImageUploadPolicy] = None,
                     logo: Optional[str] = None,
                     is_censored: Optional[bool] = None,
                     is_private: Optional[bool] = None,
                     anime_ids: Optional[List[int]] = None,
                     manga_ids: Optional[List[int]] = None,
                     ranobe_ids: Optional[List[int]] = None,
                     character_ids: Optional[List[int]] = None,
                     club_ids: Optional[List[int]] = None,
                     admin_ids: Optional[List[int]] = None,
                     collection_ids: Optional[List[int]] = None,
                     banned_user_ids: Optional[List[int]] = None):
        """Update info/settings about/of club.

        :param club_id: Club ID to modify/update
        :type club_id: int

        :param name: New name of club
        :type name: Optional[str]

        :param join_policy: New join policy of club
        :type join_policy: Optional[JoinPolicy]

        :param description: New description of club
        :type description: Optional[str]

        :param display_images: New display images status of club
        :type display_images: Optional[bool]

        :param comment_policy: New comment policy of club
        :type comment_policy: Optional[CommentPolicy]

        :param topic_policy: New topic policy of club
        :type topic_policy: Optional[TopicPolicy]

        :param page_policy: New page policy of club
        :type page_policy: Optional[PagePolicy]

        :param image_upload_policy: New image upload policy of club
        :type image_upload_policy: Optional[ImageUploadPolicy]

        :param logo: Path to new image for club logo
        :type logo: Optional[str]

        :param is_censored: New censored status of club
        :type is_censored: Optional[bool]

        :param is_private: New privacy status of club
        :type is_private: Optional[bool]

        :param anime_ids: New anime ids of club
        :type anime_ids: Optional[List[int]]

        :param manga_ids: New manga ids of club
        :type manga_ids: Optional[List[int]]

        :param ranobe_ids: New ranobe ids of club
        :type ranobe_ids: Optional[List[int]]

        :param character_ids: New character ids of club
        :type character_ids: Optional[List[int]]

        :param club_ids: New club ids of club
        :type club_ids: Optional[List[int]]

        :param admin_ids: New admin ids of club
        :type admin_ids: Optional[List[int]]

        :param collection_ids: New collection ids of club
        :type collection_ids: Optional[List[int]]

        :param banned_user_ids: New banned user ids of club
        :type banned_user_ids: Optional[List[int]]

        :return: Updated club info
        :rtype: Optional[Club]
        """
        image_data = await Utils.get_image_data(logo)
        data_dict = Utils.create_data_dict(
            dict_name=DICT_NAME,
            name=name,
            join_policy=join_policy,
            description=description,
            display_images=display_images,
            comment_policy=comment_policy,
            topic_policy=topic_policy,
            page_policy=page_policy,
            image_upload_policy=image_upload_policy,
            logo=image_data,
            is_censored=is_censored,
            is_private=is_private,
            anime_ids=anime_ids,
            manga_ids=manga_ids,
            ranobe_ids=ranobe_ids,
            character_ids=character_ids,
            club_ids=club_ids,
            admin_ids=admin_ids,
            collection_ids=collection_ids,
            banned_user_ids=banned_user_ids)
        form_data = Utils.create_form_data(data_dict)

        response = await self._client.request(
            self._client.endpoints.club(club_id),
            form_data=form_data,
            request_type=RequestType.PATCH)

        return Utils.validate_response_data(cast(Dict[str, Any], response),
                                            data_model=Club)

    @method_endpoint('/api/clubs/:id/animes')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def animes(self, club_id: int):
        """Returns anime list of club.

        :param club_id: Club ID to get anime list
        :type club_id: int

        :return: Club's anime list
        :rtype: List[AnimeInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_animes(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=AnimeInfo)

    @method_endpoint('/api/clubs/:id/mangas')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def mangas(self, club_id: int):
        """Returns manga list of club.

        :param club_id: Club ID to get manga list
        :type club_id: int

        :return: Club's manga list
        :rtype: List[MangaInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_mangas(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=MangaInfo)

    @method_endpoint('/api/clubs/:id/ranobe')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def ranobe(self, club_id: int):
        """Returns ranobe list of club.

        :param club_id: Club ID to get ranobe list
        :type club_id: int

        :return: Club's ranobe list
        :rtype: List[RanobeInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_ranobe(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=RanobeInfo)

    @method_endpoint('/api/clubs/:id/characters')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def characters(self, club_id: int):
        """Returns character list of club.

        :param club_id: Club ID to get character list
        :type club_id: int

        :return: Club's character list
        :rtype: List[CharacterInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_characters(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=CharacterInfo)

    @method_endpoint('/api/clubs/:id/collections')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def collections(self, club_id: int):
        """Returns collection list of club.

        :param club_id: Club ID to get collection list
        :type club_id: int

        :return: Club's collection list
        :rtype: List[Topic]
        """
        response = await self._client.request(
            self._client.endpoints.club_collections(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=Topic)

    @method_endpoint('/api/clubs/:id/clubs')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def clubs(self, club_id: int):
        """Returns clubs info list of club.

        :param club_id: Club ID to get clubs info list
        :type club_id: int

        :return: Club's clubs info list
        :rtype: List[ClubInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_clubs(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=ClubInfo)

    @method_endpoint('/api/clubs/:id/members')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def members(self, club_id: int):
        """Returns member list of club.

        :param club_id: Club ID to get member list
        :type club_id: int

        :return: Club's member list
        :rtype: List[UserInfo]
        """
        response = await self._client.request(
            self._client.endpoints.club_members(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=UserInfo)

    @method_endpoint('/api/clubs/:id/images')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=[])
    async def images(self, club_id: int):
        """Returns images of club.

        :param club_id: Club ID to get images
        :type club_id: int

        :return: Club's image list
        :rtype: List[ClubImage]
        """
        response = await self._client.request(
            self._client.endpoints.club_images(club_id))

        return Utils.validate_response_data(cast(List[Dict[str, Any]],
                                                 response),
                                            data_model=ClubImage)

    @method_endpoint('/api/clubs/:id/join')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def join(self, club_id: int):
        """Joins club by ID.

        When trying to join already joined club,
        method will log out 403 error and return False

        :param club_id: Club ID to join
        :type club_id: int

        :return: Status of join
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.club_join(club_id),
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)

    @method_endpoint('/api/clubs/:id/leave')
    @exceptions_handler(ShikimoriAPIResponseError, fallback=False)
    async def leave(self, club_id: int):
        """Leaves club by ID.

        When trying to leave already left club,
        method will log out 403 error and return False

        :param club_id: Club ID to leave
        :type club_id: int

        :return: Status of leave
        :rtype: bool
        """
        response = await self._client.request(
            self._client.endpoints.club_leave(club_id),
            request_type=RequestType.POST)

        return Utils.validate_response_code(cast(int, response),
                                            check_code=ResponseCode.SUCCESS)
